// ACE notebook support
// version 0.2.19.2024

"use strict"; // nb. soft tabs 4
/* globals editor: false, MathJax: false */

/*
    adds notebook content to ACE editor
    requires the modified ACE editor (currently)

    any line with special format will have content inserted below it
        OR
    insert() can be called by user to insert content programmatically

    format:
        /// notebook ID LINES NTYPE SRC
            eg.
        /// notebook fig1 10 img image.png

    where:
        id      is a unique HTML document element id
        lines   is an integer number of lines to reserve, typically 3 to 20
        ntype   is one of:  img  html  tex
        src     depends on ntype:

                ntype       src
                -----       ---
                img         imagefile           image file is read and inserted
                html        htmlcode            html gets inserted
                tex         tex string          for MathJax *

    * MathJax must be loaded by page if 'tex' is used, eg:
    <script src="https://polyfill.io/v3/polyfill.min.js?features=es6"></script>
    <script id="MathJax-script" src="https://cdn.jsdelivr.net/npm/mathjax@3/es5/tex-chtml.js"></script>

*/

// error;

let ace_nb = {};

ace_nb.init = () => {

let _                   = ace_nb,   // singleton (currently)
    logC                = 1,        // number for log lines
    ic                  = 0,        // initialization flag
    nbEnableContent     = 1,        // if 0 then no content is added below outer div
    pgmContent          = {},       // [id]  HTML for entry from programmed source
    nbLineHt            = 0,        // computed line height from ACE
    nbWrap              = {},       // [id] computed wrap
    nbLayer             = null,     // top text layer
    nbCache             = {},       // [id]  cache
    nbLineVis           = {},       // [row] id, row  for folding
    nbLineTrack         = {},       // [currentRow] cache
    nbTOC               = {};       // [id] row  for table of contents

_.dbg       = 0;        // debug flag

_.nbEnable  = 1;        // notebook modifications in ACE enabled

_.nbMarker  = '/// notebook'; // marker string for notebook insertion

_.nbBorders = 0;        // show borders around inserted content (debugging)

function log(m)         {
    if (!_.dbg) return;
    let el = document.getElementById('log');
    if (el) el.value += ((logC++).toString())+' '+m+'\n';
}

function log2(m)        {
    if (!_.dbg) return;
    log(m);
}

function logCon(m)      { if (_.dbg) console.log(((logC++).toString())+' '+m);  }
_.log = log;

function log2Con(m)     { console.log(((logC++).toString())+' '+m);  }
_.log2 = log2;          // used in ACE trace

function jstr(e)        { return JSON.stringify(e); }
_.jstr = jstr;

function dge(id)        { return document.getElementById(id); }
_.dge = dge;

function dce(tag)       { return document.createElement(tag); }
_.dce = dce;

function dac(el,c)      { return el.appendChild(c); }
_.dac = dac;

function addCSS(s)      { document.head.insertAdjacentHTML("beforeend", s ); }
_.addCSS = addCSS;


// called by user:
// ---------------

// set fn to draw dynamic content, eg. plots, d3 etc.
function draw(id,fn) {
    log('nbRedraw set fn  id = '+id);
    pgmContent[id] = fn;
}
_.draw = draw;

// insert content into editor  line must have at least blank and \n
function insert(row,id='',lines=10,ntype='',src='') {
    row = +row; row--;  // make numeric and origin zero
    log('inserted row '+row+' id '+id);
    nbLineTrack[row] = {row:row,id:id,lnC:lines,ntype:ntype,src:src,top:0,ht:0,w:0};
    log('insert nbLineTrack = '+jstr(nbLineTrack));
    nbTOC[id] = row;
}
_.insert = insert;

// scroll editor to row that matches the id
// useful in a onclick for <a> or <button>
function goto(id) {
    let row = nbTOC[id] ?? null;
    log('goto  '+jstr({id:id,row:row}));
    if (row!=null) editor.gotoLine(row);
}
_.goto = goto;

// fns below are called by ACE
// ---------------------------

function isNotebook(row) {
    // log('isNotebook');
    let r = !!nbLineTrack[row], ln, r0=0;
    if (!r) {
        // log('  checking row '+row);
        ln = editor.session.doc.$lines[row] ?? '';
        r0 = ln.indexOf(_.nbMarker) >= 0;
        // if (r0) log('  NEW notebook line');
        r = r0;
    }
    // log('isNotebook  row = '+row+'  r = '+r);
    return r;
}
_.isNotebook = isNotebook;

function parseLine(row) {
    // nb. see         var lines = this.doc.getAllLines();  for OOP access  zvzv
    let ln = editor.session.doc.$lines[row] ?? '';  // zvzv make this OOP

    // log('  parseLine  ln = '+ln);

    if ( ln.indexOf(_.nbMarker) < 0 ) return null;

    // nb. we could cache this but then it would not be dynamic
    //   there aren't too many nb lines in a document typically

    // control line format eg.
    // 0 1        2    3  4
    ///  notebook fig1 10 img image.png
    //                    [anytext] /// notebook id     lines   ntype   src
    let m = ln.match(/.*\/\/\/\s+\S+\s+(\S+)\s+(\S+)\s*(\S*)\s*(.*)/);
    // let m = ln.match(/\s*\S+\s+\S+\s+(\S+)\s+(\S+)\s*(\S*)\s*(.*)/);
    m = m ?? [];
    // log('    m = '+jstr(m));
    let id      = m[1] || '',
        lnC     = m[2] || 10,
        ntype   = m[3] || '',
        src     = m[4] || '';

    if (src=='') return null;
    return {id:id,lnC:lnC,ntype:ntype,src:src,top:0,ht:0,w:0};

}

function getLineInfo(row) {
    let r = nbLineTrack[row];
    if (!r) r = parseLine(row);
    // nb. we do not set nbLineTrack[row] after parseLine
    // in order to keep changes in the editor dynamic
    // log('  getLineInfo  row = '+row+'  got: '+jstr(r));
    return r;
}

function checkLineChg(row) {
    log('checkLineChg row = '+row);
    if (!nbLineTrack[row]) return;
    delete nbLineTrack[row];
    clearCache();
    updateACE();
}

function edInsert(position, text) {
    log('edInsert  pos = '+jstr(position)+' text = '+jstr(text));
    let n;  // count nr of lines inserted
    for (let i=n=0; i<text.length; i++) n += +('\n'===text[i]);
    log('  n = '+n);
    let row = +position.row;
    if (n == 0) {
        checkLineChg(row); return; }
    let b, c, clr = 0, m;
    log('  row = '+row);
    log('  0 nbLineTrack = '+jstr(nbLineTrack));
    for (b in nbLineTrack) {
        b = +b;
        if (b>=row) {
            c = nbLineTrack[b];
            delete nbLineTrack[b];
            m=b+n;
            nbLineTrack[m] = c;
            nbTOC[c.id] = m;
            clr = 1;
        }
    }
    log('  1 nbLineTrack = '+jstr(nbLineTrack));
    if (clr) { clearCache(); updateACE(); }
}
_.edInsert = edInsert;

function edRemove(range) {
    log('edRemove  range = '+jstr(range));
    let r1 = range.start.row, r2 = range.end.row, n = r2-r1;
    log('  '+jstr({r1:r1,r2:r2,n:n}));
    let row = r2, b, c, clr = 0, m;
    log('  0 nbLineTrack = '+jstr(nbLineTrack));
    if (n==0) checkLineChg(r1);
    if (n<=0) return;
    for (b in nbLineTrack) {
        b = +b;
        if (b>=row) {
            c = nbLineTrack[b];
            delete nbLineTrack[b];
            m = b-n;
            if (m>=0) {
                nbLineTrack[m] = c;
                nbTOC[c.id] = m;
            }
            clr = 1;
        }
    }
    log('  1 nbLineTrack = '+jstr(nbLineTrack));
    if (clr) { clearCache(); updateACE(); }
}
_.edRemove = edRemove;

function nbInit() {
    let p, c;
    log('initializing');

    // editor.renderer.on("afterRender", _.nbUpdScroll);  // zvzv test

    nbLineHt = editor.renderer.layerConfig.lineHeight;
    // log('  nbLineHt = '+nbLineHt);

    p = editor.renderer.content;        // get content layer

    c = document.createElement('div');
    c.classList.add('nbLayer');
    c.style.margin  = '0px 4px';

    p.appendChild(c);           // text layer

    nbLayer = c;

    // log('  nbLayer.outerHTML = '+nbLayer.outerHTML);
    addCSS(
`<style>
.nbLayer {
    height: 100000px;
    width: 100000px;
    position: absolute;
    overflow: hidden;
}
.nbImg {
    pointer-events: none;       // turn "visual search" in Edge off
    overflow: hidden;
}
</style>
`
);
    // z-index: 10;
    ic++;
    // log('init done');
}

function updateACE()
{
    let er = editor.renderer;       // zvzv make this OOP
    er.$loop.schedule(er.CHANGE_FULL);
    log('updateACE');
    // nb. this.$loop.schedule(this.CHANGE_FULL);
}

function nbUpdScroll() {
    // shift nb layers to match ACE layers for scrolling
    if (!_.nbEnable) return;

    log('nbUpdScroll');

    if (ic==0) nbInit();

    let txt = editor.renderer.$textLayer.element;       // get text layer

    if (txt && nbLayer) {
        let xfm, top;

        // move text layer
        xfm = txt.style.transform; top = txt.style.top;
        log('  xfm = '+jstr(xfm)+'  top = '+jstr(top));
        if (xfm)        nbLayer.style.transform = xfm;
        else if (top)   nbLayer.style.top = top;
        else log('  no upd');
    } else log('info: missing txt or nbLayer');
}
_.nbUpdScroll = nbUpdScroll;

function clearCache() {
    // clear cache
    log('clear cache');
    for (const id in nbCache) deleteLine(id);
    nbLineVis = {};
}

function nbExpandFold() {
    clearCache();
}
_.nbExpandFold = nbExpandFold;

function nbAddFold() {
    clearCache();
}
_.nbAddFold = nbAddFold;

function nbUpdVis(show,start,end) {     // handle scrolling and folded divs
    if (!_.nbEnable) return;

    // called by ACE to update folding
    // log('nbUpdVis  show = '+show+'  range: '+start+' to '+end);
    // if folded: display = none

    if (ic==0) nbInit();

    // hide other lines if not shown
    if (show) return;

    // hide folded divs
    for (let i=start; i<end; i++) {
        if ( nbLineVis[i] ) {
            let id = nbLineVis[i];
            // log('  hide 1 row '+i+' id = '+id);
            let c = dge('nb_'+id);
            if (!c) {
                log('error: cannot get id '+id);
            } else {
                c.style.display = 'none';       // hide el
            }
        }
    }
}
_.nbUpdVis = nbUpdVis;

function deleteLine(id) {
    if (!nbCache[id]) return;
    log('  deleting '+ id);
    let id2 = 'nb_'+id, c = dge(id2);
    if (c) nbLayer.removeChild(c);
    log('  deleted id '+id);
    delete nbCache[id];     // remove cache entry
    // log('  nbLayer.outerHTML = '+nbLayer.outerHTML);
}

function setWrap(row,compWrap) {
    log('setWrap  row = '+row+'  computed wrap = '+jstr(compWrap));
    // setWrap is called first by ACE
    // return computed wrap with modifed wrap appended to the array
    // cache mgmt from editor content changes are done here

    // isNotebook already returned true

    let newLine = getLineInfo(row);
    if (!newLine) {
        log('error: getLineInfo failed');
        clearCache(); return compWrap;  }

    let r, prevW = 0, id = newLine.id, upd = 0;
    prevW = nbWrap[id] || null;
    if (!compWrap) compWrap = [];
    nbWrap[id] = compWrap.length;                   // store computed wrap for div sizing
    log('  nbWrap = '+jstr(nbWrap));

    let ln = editor.session.doc.$lines[row] ?? '';  // zvzv make this OOP
    ln = ln.length+1;
    r = compWrap;                                   // get computed wrap
    for (let i=0;i<newLine.lnC;i++) r.push(ln);     // append repeated fake wraps to reserve space for content
    log('  r = '+jstr(r));

    // nb. if this item grew, other items may not be visible now
    // if ( prevW && prevW != compWrap ) {   // reset for layout
    //     // debugger;
    //     nbCache[id].top = 0;
    //     // nbCache[id].ht = 0;
    // }

    return r;

/*  NIU
    id = nbLineVis[row] || null;                    // reset id to visible lines
    log('  nbLineVis[row]  row = '+row+' id = '+id);
    if (!id) return r;

    if (id != newLine.id) {                         // id changed
        log('  id chg');
        if ( nbCache[id] ) {                        // delete old cache and content
            log('  delete  id = '+id);
            deleteLine(id);
            upd = 1;
        } else if (id) {
            id = newLine.id;
            nbCache[id] = newLine;                  // update cache
            log('  add cacheLine  id = '+id+' '+jstr(newLine));
            upd = 1;
        }
    } else {                                        // same id, check for other changes from content
        let cacheLine = nbCache[id];
        if (cacheLine) {
            if ( cacheLine.ntype != newLine.ntype ||
                 cacheLine.src != newLine.src ) {
                log('  ntype or src chg  delete id = '+id);
                // deleteLine(id);
                nbCache[id] = newLine;      // update cache
                upd = 1;
            }
            if ( prevW != compWrap ) {
                log('  wrap chg  id = '+id);
                // clearCache();               // clear all
                // nb. if this item grew, other items may not be visible now
                upd = 1;
            }
        } else log('  info: no cacheLine');
    }

    if (upd) updateACE();

    return r;
*/
}
_.setWrap = setWrap;

function setLineVis(row) {                  // set visible line
    // log('setLineVis  row = '+row);
    let t = getLineInfo(row), id = t.id;
    // log('  showing id = '+id);
    if (id != '') nbLineVis[row] = id;      // line visible
    // log('  nbLineVis = '+jstr(nbLineVis));
}
_.setLineVis = setLineVis;

function refreshPlot(id,c,createFlag) {
    let p = pgmContent[id];
    if (p && (typeof p == 'function')) {
        log('refreshPlot calling fn for '+id);
        let e = nbCache[id];
        p({c:c,id:id,src:e.src,createFlag:createFlag});
        return 1;   // was called
    }
    return 0;       // not called
}

function setDiv(id,c) {
    let wrap = nbWrap[id] || 0,
        adj = nbLineHt * (wrap+1);
    let e = nbCache[id];
    e.top = (parseInt(e.top) + adj)+'px';
    c.style.top = e.top;
    // let w = editor.renderer.content.style.width;
    // c.style.width = e.w;   // nb. browser will left justify, if this is off

    let h = parseInt(e.ht);
    if (h == 0) h = e.lnC * nbLineHt;
    else h = h - adj;
    e.h = h+'px';
    c.style.height = e.h;
    log('setDiv  '+jstr({wrap:wrap,adj:adj,top:e.top,h:e.h,w:e.w}));
}

function setContent(id,c) {
    let e = nbCache[id];
    if ( e.ntype == 'img' )
        c.innerHTML = '<img id="'+id+'" class="nbImg" height='+e.h+' src="'+e.src+'">';
    else if ( e.ntype == 'html' )     c.innerHTML = e.src;
    else if ( e.ntype == 'tex' )      c = setMathJax(c,e.src);
    // else if ( ntype == 'amath' )    c = setAsciiMath(c,src); // NFN
    else c.innerHTML = '<div id="'+id+'" style="width:100%;height:100%;"></div>';
    // style="width:'+w+';height:'+h+'"
}

function setMathJax(c,src) {
    if (!window.MathJax) { log('error: got MathJax content, but pkg not loaded'); return; }
    // see: view-source:https://mathjax.github.io/MathJax-demos-web/input-tex2chtml.html
    MathJax.texReset();
    let options = MathJax.getMetricsFor(c);
    options.display = 1;
    MathJax.tex2chtmlPromise(src, options).then(function (node) {
        c.appendChild(node);
        MathJax.startup.document.clear();
        MathJax.startup.document.updateDocument();  // upd doc for adjusted CSS for new equation
    });
}

function nbUpd(parent,row) {
    if (!_.nbEnable) return;

    log('nbUpd  row = '+row);

    if (!parent) { log('error: no parent'); return; }

    let newLine = getLineInfo(row), id = newLine.id || '';

    if (!newLine) return;
    if (id =='') { log('info: no id in line'); return; }

    newLine.top = parent.style.top;
    // newLine.w = parent.style.width;
    newLine.ht = parent.style.height;

    log('  newLine   = '+jstr(newLine));

    let c = dge('nb_'+id), cacheLine = nbCache[id] || null;

    log('  cacheLine = '+jstr(cacheLine));

    nbCache[id] = newLine;      // update also

    if (!c) {                       // no el found
        log('  new el nb_'+id);

        // if (dge(id)) { log('error: duplicate element: '+id); return; }

        // if (!cacheLine)
        nbCache[id] = newLine;      // update also

        // create el
        let c = document.createElement('div');
        // c.style.backgroundColor = "#c84";        // test
        if (_.nbBorders) c.style.border = 'solid 1px #ccc';
        c.id = 'nb_'+id;
        c.style.overflow = "hidden";
        c.style.position = 'absolute';
        // c.style.zIndex = 10;
        setDiv(id,c);
        if (nbEnableContent) {
            setContent(id,c); refreshPlot(id,c,1);
        }
        nbLayer.appendChild(c);
        // log('  NEW nbLayer.outerHTML = '+nbLayer.outerHTML);

        // updateACE();

    } else {

        c.style.display = 'block';  // show el

        if (    newLine.lnC     != cacheLine.lnC    ||
                newLine.ntype   != cacheLine.ntype  ||
                newLine.src     != cacheLine.src
                ) {
            log('  update');
            deleteLine(id);
            updateACE();
        } else if ( newLine.top != cacheLine.top ||
                    newLine.ht != cacheLine.ht      ) {
            setDiv(id,c);
            nbCache[id] = newLine;
            // cacheLine.top = newLine.top; cacheLine.ht = newLine.ht;
            if (nbEnableContent) refreshPlot(id,c,newLine.src,0);
            // log('  UPD nbLayer.outerHTML = '+nbLayer.outerHTML);
        } else log('  no change');
    }
}
_.nbUpd = nbUpd;

};  // end init

ace_nb.init();

// end
